@@ -24,12 +24,10 @@ from os.path import isfile
 from collections import OrderedDict
 from contextlib import contextmanager
 
-# Workaround to support the correct input for both Python 2 and 3. Always use
-# input() which will point to the correct builtin.
-try:
+import sys
+if sys.version_info[0] < 3:
     input = raw_input
-except NameError:
-    pass
+    PermissionError = RuntimeError
 
 
 class FeatureGroup(click.Group):
@@ -48,7 +46,7 @@ class FeatureGroup(click.Group):
             ctx.obj.api = Connection.from_file('~/.transcriptic')
         except (FileNotFoundError, OSError):
             # This defaults to feature_groups = []
-            ctx.obj.api = Connection(use_environ=False)
+            ctx.obj.api = Connection()
 
         rows = []
         for subcommand in self.list_commands(ctx):
@@ -80,6 +78,24 @@ class FeatureCommand(click.Command):
         self.feature = feature
 
 
+class HiddenOption(click.Option):
+    """Monkey patch of click Option to enable hidden options
+    TODO: Deprecate once Click 7 lands and use `hidden` option instead
+    """
+    def __init__(self, *param_decls, **attrs):
+        __hidden__ = attrs.pop('hidden', True)
+        click.Option.__init__(self, *param_decls, **attrs)
+        self.__hidden__ = __hidden__
+
+    def get_help_record(self, ctx):
+        """This hijacks the help record so that a hidden option does not show 
+        up in the help text
+        """
+        if self.__hidden__:
+            return
+        click.Option.get_help_record(self, ctx)
+
+
 class ContextObject(object):
     """Object passed along Click context
     Note: `ctx` is passed along whenever the @click.pass_context decorator is
@@ -101,39 +117,55 @@ _CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])
 
 
 @click.group(context_settings=_CONTEXT_SETTINGS, cls=FeatureGroup)
-@click.option('--apiroot', default=None)
+@click.option('--api-root', default=None, hidden=True, cls=HiddenOption)
+@click.option('--email', default=None, hidden=True, cls=HiddenOption)
+@click.option('--token', default=None, hidden=True, cls=HiddenOption)
+@click.option('--organization', '-o', default=None, hidden=True, cls=HiddenOption)
 @click.option(
     '--config',
     envvar='TRANSCRIPTIC_CONFIG',
     default='~/.transcriptic',
     help='Specify a configuration file.'
 )
-@click.option('--organization', '-o', default=None)
 @click.version_option(prog_name="Transcriptic Python Library (TxPy)")
 @click.pass_context
-def cli(ctx, apiroot, config, organization):
+def cli(ctx, api_root, email, token, organization, config):
     """A command line tool for working with Transcriptic.
-    Note: This is the main entry point of the CLI. ALl commands go through
-    this `group` before calling a sub-`command`
+    
+    Note: This is the main entry point of the CLI. If specifying credentials,
+    note that the order of preference is: --flag, environment then config file.
+    
+    Example: `transcriptic --organization "my_org" projects` >> 
+    `export USER_ORGANIZATION="my_org"` >> `"organization_id": "my_org" in ~/.transcriptic
     """
     if ctx.invoked_subcommand in ['login', 'compile', 'preview', 'summarize', 'init']:
         # For login/local commands, initialize empty connection
         ctx.obj = ContextObject()
-        ctx.obj.api = Connection(use_environ=False)
+        ctx.obj.api = Connection()
     else:
         try:
             ctx.obj = ContextObject()
-            ctx.obj.api = Connection.from_file(config)
-            if organization is not None:
-                ctx.obj.api.organization_id = organization
-            if apiroot is not None:
-                ctx.obj.api.api_root = apiroot
-        except:
+            api = Connection.from_file(config)
+            api.api_root = (
+                api_root or os.environ.get('BASE_URL', None) or api.api_root
+            )
+            api.organization_id = (
+                organization or os.environ.get('USER_ORGANIZATION', None) or
+                api.organization_id
+            )
+            api.email = (
+                email or os.environ.get('USER_EMAIL', None) or api.email
+            )
+            api.token = (
+                token or os.environ.get('USER_TOKEN', None) or api.token
+            )
+            ctx.obj.api = api
+        except (OSError, IOError):
             click.echo("Welcome to TxPy! It seems like your `.transcriptic` config file is missing or out of date")
             analytics = click.confirm("Send TxPy CLI usage information to improve the CLI user "
                                       "experience?", default=True)
-            ctx.obj.api = Connection(use_environ=False)  # Initialize empty connection
-            ctx.invoke(login, analytics=analytics)
+            ctx.obj.api = Connection()  # Initialize empty connection
+            ctx.invoke(login, api_root=api_root, analytics=analytics)
     if ctx.obj.api.analytics:
         try:
             ctx.obj.api._post_analytics(event_action=ctx.invoked_subcommand, event_category="cli")
@@ -249,29 +281,16 @@ def upload_release(ctx, archive, package):
     with click.progressbar(None, 100, "Upload Progress",
                            show_eta=False, width=70,
                            fill_char="|", empty_char="-") as bar:
-        bar.update(10)
-        info = ctx.obj.api.get(ctx.obj.api.get_route(
-            'upload_sign'), params={'name': archive})
-        bar.update(30)
-        aws_url = ctx.obj.api.get_route('aws_upload')
-        files = {'file': open(os.path.basename(archive), 'rb')}
-        data = OrderedDict([
-            ('key', info['key']),
-            ('AWSAccessKeyId', 'AKIAJVJ67EJYCQXO7ZSQ'),
-            ('acl', 'private'),
-            ('success_action_status', '201'),
-            ('policy', info['policy']),
-            ('signature', info['signature']),
-        ])
-        response = ctx.obj.api.post(aws_url, data=data, files=files, headers={},
-                                    status_response={'201': lambda resp: resp})
+        bar.update(40)
+        file = open(os.path.basename(archive), 'rb')
+        url = ctx.obj.api.upload_to_uri(
+            file, 'application/zip, application/octet-stream', archive, archive
+        )
         bar.update(20)
-        response_tree = ET.fromstring(response.content)
-        loc = dict((i.tag, i.text) for i in response_tree)
         try:
             up = ctx.obj.api.post_release(
                 data=json.dumps({"release":
-                                 {"binary_attachment_url": loc["Key"]}}
+                                 {"binary_attachment_url": url}}
                                 ),
                 package_id=package_id
             )
@@ -308,6 +327,45 @@ def upload_release(ctx, archive, package):
                                                                 package_id))
 
 
+@cli.command('upload-dataset')
+@click.argument('file_path', type=click.Path(exists=True),
+                metavar="FILE")
+@click.argument('title', metavar="TITLE")
+@click.argument('run_id', metavar="RUN-ID")
+@click.option('--tool', '-t', required=True,
+              help="Name of analysis tool used for generating the dataset")
+@click.option('--version', '-v', required=True,
+              help="Version of analysis tool used for generating the dataset")
+@click.pass_context
+def upload_dataset(ctx, file_path, title, run_id, tool, version):
+    """Uploads specified file as an analysis dataset to the specified run."""
+    resp = ctx.obj.api.upload_dataset_from_filepath(
+        file_path=file_path,
+        title=title,
+        run_id=run_id,
+        analysis_tool=tool,
+        analysis_tool_version=version
+    )
+    try:
+        data_id = resp['data']['id']
+        run_route = ctx.obj.api.url(
+            "/api/runs/{}?fields[runs]=project_id".format(run_id)
+        )
+        run_resp = ctx.obj.api.get(run_route)
+        project = run_resp['data']['attributes']['project_id']
+        data_url = "{}/analysis/{}".format(
+            ctx.obj.api.get_route(
+                'datasets',
+                project_id=project,
+                run_id=run_id
+            ),
+            data_id
+        )
+        click.echo("Dataset uploaded to {}".format(data_url))
+    except KeyError:
+        click.echo("An unexpected response was returned from the server. ")
+
+
 @cli.command()
 @click.pass_context
 @click.option(
@@ -456,14 +514,14 @@ def projects(ctx, i, json_flag):
     """List the projects in your organization"""
     try:
         projects = ctx.obj.api.projects()
-        proj_names = {}
+        proj_id_names = {}
         all_proj = {}
         for proj in projects:
             status = " (archived)" if proj['archived_at'] else ""
-            proj_names[proj["name"]] = proj["id"]
-            all_proj[proj["name"] + status] = proj["id"]
+            proj_id_names[proj["id"]] = proj["name"]
+            all_proj[proj["id"]] = proj["name"] + status
         if i:
-            return proj_names
+            return proj_id_names
         elif json_flag:
             return click.echo(json.dumps(projects))
         else:
@@ -471,9 +529,9 @@ def projects(ctx, i, json_flag):
             click.echo('{:^40}'.format("PROJECT NAME") + "|" +
                        '{:^40}'.format("PROJECT ID"))
             click.echo('{:-^80}'.format(''))
-            for name, i in list(all_proj.items()):
+            for proj_id, name in list(all_proj.items()):
                 click.echo('{:<40}'.format(name) + "|" +
-                           '{:^40}'.format(i))
+                           '{:^40}'.format(proj_id))
                 click.echo('{:-^80}'.format(''))
     except RuntimeError:
         click.echo("There was an error listing the projects in your "
@@ -510,7 +568,7 @@ def runs(ctx, project_name, json_flag):
         else:
             click.echo(
                 '\n{:^120}'.format("Runs in Project '%s':\n" %
-                                   get_project_name(project_name))
+                                   get_project_name(project_id))
             )
             click.echo('{:^30}'.format("RUN TITLE") + "|" +
                        '{:^30}'.format("RUN ID") + "|" +
@@ -535,10 +593,13 @@ def create_project(ctx, name, dev):
     existing = ctx.obj.api.projects()
     for p in existing:
         if name == p['name'].split('.')[-1]:
-            click.echo("You already have an existing project with the name "
-                       "\"%s\".  Please choose a different project name." %
-                       name)
-            return
+            click.confirm(
+                "You already have an existing project with the name '{}'. "
+                "Are you sure you want to create another one?".format(name),
+                default=False,
+                abort=True
+            )
+            break
     try:
         new_proj = ctx.obj.api.create_project(name)
         click.echo(
@@ -1171,6 +1232,20 @@ def _get_quick_launch(ctx, protocol, project):
 def select_org(ctx):
     """Allows you to switch organizations"""
     org_list = [{"name": org['name'], "subdomain": org['subdomain']} for org in ctx.obj.api.organizations()]
+    organization = org_prompt(org_list)
+
+    r = ctx.obj.api.get_organization(org_id=organization)
+    if r.status_code != 200:
+        click.echo("Error accessing organization: %s" % r.text)
+        sys.exit(1)
+
+    ctx.obj.api.organization_id = organization
+    ctx.obj.api.save(ctx.parent.params['config'])
+    click.echo('Logged in with organization: {}'.format(organization))
+
+
+def org_prompt(org_list):
+    """Organization prompt for helping with selecting organization"""
     if len(org_list) < 1:
         click.echo("Error: You don't appear to belong to any organizations. \n"
                    "Visit {} and create an organization.".format('https://secure.transcriptic.com'))
@@ -1186,10 +1261,13 @@ def select_org(ctx):
         def parse_valid_org(indx):
             from click.exceptions import BadParameter
             try:
-                return org_list[int(indx) - 1]['subdomain']
+                org_indx = int(indx) - 1
+                if org_indx < 0 or org_indx >= len(org_list):
+                    raise ValueError("Value out of range")
+                return org_list[org_indx]['subdomain']
             except:
                 raise BadParameter("Please enter an integer between 1 and %s" %
-                                   (len(org_list)), ctx=ctx)
+                                   (len(org_list)))
 
         organization = click.prompt(
             'Which organization would you like to log in as',
@@ -1197,37 +1275,40 @@ def select_org(ctx):
             prompt_suffix='? ', type=int,
             value_proc=lambda x: parse_valid_org(x)
         )
-
-    r = ctx.obj.api.get_organization(org_id=organization)
-    if r.status_code != 200:
-        click.echo("Error accessing organization: %s" % r.text)
-        sys.exit(1)
-
-    ctx.obj.api.organization_id = organization
-    ctx.obj.api.save(ctx.parent.params['config'])
-    click.echo('Logged in with organization: {}'.format(organization))
+        # Catch since `value_proc` doesn't properly parse default
+        if organization == 1:
+            organization = org_list[0]['subdomain']
+    return organization
 
 
 @cli.command()
-@click.option('--api-root', default='https://secure.transcriptic.com')
 @click.pass_context
-def login(ctx, api_root, analytics=True):
+def login(ctx, api_root=None, analytics=True):
     """Authenticate to your Transcriptic account."""
+    # For logging in, we should default to the transcriptic domain
+    if api_root is None:
+        api_root = "https://secure.transcriptic.com"
     email = click.prompt('Email')
     password = click.prompt('Password', hide_input=True)
-    r = ctx.obj.api.post(routes.login(api_root=api_root), data=json.dumps({
-        'user': {
-            'email': email,
-            'password': password,
+    r = ctx.obj.api.post(
+        routes.login(api_root=api_root),
+        data=json.dumps({
+            'user': {
+                'email': email,
+                'password': password,
+            },
+        }),
+        headers={
+            'Accept': 'application/json',
+            'Content-Type': 'application/json',
         },
-    }), headers={
-        'Accept': 'application/json',
-        'Content-Type': 'application/json',
-    }, status_response={
-        '200': lambda resp: resp,
-        'default': lambda resp: resp
-    },
-        custom_request=False)
+        status_response={
+            '200': lambda resp: resp,
+            '401': lambda resp: resp,
+            'default': lambda resp: resp
+        },
+        custom_request=False
+    )
     if r.status_code != 200:
         click.echo("Error logging into Transcriptic: %s" % r.json()['error'])
         sys.exit(1)
@@ -1236,32 +1317,7 @@ def login(ctx, api_root, analytics=True):
              user['test_mode_authentication_token'])
     user_id = user.get("id")
     feature_groups = user.get('feature_groups')
-    if len(user['organizations']) < 1:
-        click.echo("Error: You don't appear to belong to any organizations. \n"
-                   "Visit %s and create an organization." % api_root)
-        sys.exit(1)
-    if len(user['organizations']) == 1:
-        organization = user['organizations'][0]['subdomain']
-    else:
-        click.echo("You belong to %s organizations:" %
-                   len(user['organizations']))
-        for indx, o in enumerate(user['organizations']):
-            click.echo("%s.  %s (%s)" % (indx + 1, o['name'], o['subdomain']))
-
-        def parse_valid_org(indx):
-            from click.exceptions import BadParameter
-            try:
-                return user['organizations'][int(indx) - 1]['subdomain']
-            except:
-                raise BadParameter("Please enter an integer between 1 and %s" %
-                                   (len(user['organizations'])), ctx=ctx)
-
-        organization = click.prompt(
-            'Which organization would you like to log in as',
-            default=1,
-            prompt_suffix='? ', type=int,
-            value_proc=lambda x: parse_valid_org(x)
-        )
+    organization = org_prompt(user['organizations'])
 
     r = ctx.obj.api.get(routes.get_organization(api_root=api_root, org_id=organization), headers={
         'X-User-Email': email,
@@ -1286,22 +1342,30 @@ def login(ctx, api_root, analytics=True):
 @click.pass_context
 def get_project_id(ctx, name):
     projs = ctx.invoke(projects, i=True)
-    project_id = projs.get(name)
-    if not project_id:
-        project_id = name if name in list(projs.values()) else None
-        if not project_id:
+    if name in projs:
+        return name
+    else:
+        project_ids = [k for k, v in projs.items() if v == name]
+        if not project_ids:
             click.echo(
                 "The project '%s' was not found in your organization." % name)
             return
-    return project_id
+        elif len(project_ids) > 1:
+            click.echo(
+                "Found multiple projects: {} that match '{}'.".format(
+                    project_ids, name))
+            # TODO: Add project selector with dates and number of runs
+            return
+        else:
+            return project_ids[0]
 
 
 @click.pass_context
 def get_project_name(ctx, id):
-    projs = {v: k for k, v in list(ctx.invoke(projects, i=True).items())}
+    projs = ctx.invoke(projects, i=True)
     name = projs.get(id)
     if not name:
-        name = id if name in list(projs.keys()) else None
+        name = id if id in projs.values() else None
         if not name:
             click.echo(
                 "The project '%s' was not found in your organization." % name)
